{% extends 'registration/registration_base.html' %}

{% block title %}Password changed{% endblock %}

{% block registration_block %}

You have changed your password successfully.

{% endblock registration_block %}
