(function () {
  'use strict';

  angular.module('journalism_workflow.v1.editor.module', [
    'journalism_workflow.v1.editor.controllers',
    'journalism_workflow.v1.editor.directives'
  ]);

  angular.module('journalism_workflow.v1.editor.controllers', []);
  angular.module('journalism_workflow.v1.editor.directives', []);

})();
