(function () {
  'use strict';

  angular.module('test_dir.v1.s3.module', [
    'test_dir.v1.s3.controllers',
    'test_dir.v1.s3.directives'
  ]);

  angular.module('test_dir.v1.s3.controllers', []);
  angular.module('test_dir.v1.s3.directives', []);

})();
