@import 'orchestra/static/orchestra/common/css/registration.scss';
@import 'orchestra/static/orchestra/dashboard/css/dashboard.scss';
@import 'orchestra/static/orchestra/task/css/task.scss';
@import 'orchestra/static/orchestra/project_management/css/project_management.scss';
